/*****************************************************************************
*
* AVRPROG compatible boot-loader
* Version  : 0.7 (Feb. 2005)
* Compiler : avr-gcc 3.4.1 / avr-libc 1.0.2 
* size     : depends on features and startup ( minmal features < 512 words)
* by       : Martin Thomas, Kaiserslautern, Germany
*            eversmith@heizung-thomas.de
*
* License  : Copyright (c) 2005 Martin Thomas
*            Free to use. You have to mention the copyright 
*            owners in source-code and documentation of derived 
*            work. No warranty.
*
* Tested with ATmega8, ATmega16, ATmega32, ATmega128
*
* - based on the Butterfly Bootloader-Code
*   Copyright (C) 1996-1998 Atmel Corporation
*   Author(s)     : BBrandal, PKastnes, ARodland, LHM
*   The orignal code has been made available by ATMEL together with the 
*   Butterfly application code. Since ATMEL.NO had no problem with 
*   the application gcc-port they hopefully will not have any concerns about
*   publishing this port. A lot of things have been change but the ATMEL
*   "skeleton" is still in this code. Make sure to keep the copyright notice 
*   in derived work to avoid trouble.
*
* - based on boot.h from the avr-libc (c) Eric Weddington
*
****************************************************************************
*
*  The boot interrupt vector is included (this bootloader is completly in
*  ".text" section). If you need this space for further functions you have to
*  add a separate section for the bootloader-functions and add an attribute
*  for this section to _all_ function prototypes of functions in the loader. 
*  With this the interrupt vector will be placed at .0000 and the bootloader 
*  code (without interrupt vector) at the address you define in the linker
*  options for the newly created section. See the avr-libc FAQ, the avr-
*  libc's avr/boot.h documentation and the makefile for further details.
*
*  See the makefile for information how to adopt the linker-settings to 
*  the selected Boot Size (_Bxxx below)
*
*  With BOOT_SIMPLE this bootloader has 0x3DE bytes size and should fit
*  into a 512word bootloader-section.
*
*  Set AVR clock-frequency and the baudrate below, set MCU-type in 
*  makefile.
*
****************************************************************************/
/*
	Does not work reliably so far:
	- lock bits set
*/
// programmers-notepad tabsize 4
#define VERSION_HIGH '1'
#define VERSION_LOW  '0'

#include "globals.h"
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>
// function not found in boot.h to read lock/fuses
#include <util/delay.h>
#include "lowlevel.h"

/* #define turn_on_uart_transmit 	  UCSR0B |= (1<<TXEN0) //turn on UART transmit */
/* #define turn_off_uart_transmit 	  UCSR0B &= ~(1<<TXEN0) //turn off UART transmit */

#define turn_on_uart_transmit 1
#define turn_off_uart_transmit 1

/* Pin "BLPNUM" on port "BLPORT" in this port has to grounded 
   (active low) to start the bootloader */
/* #define BLPORT  PORTC */
/* #define BLDDR   DDRC */
/* #define BLPIN   PINC */
/* #define BLPNUM  PINC0 */
#define BLPORT  PORTD
#define BLDDR   DDRD
#define BLPIN   PIND
#define BLPNUM  PIND7

/* 
   Select startup-mode
   * SIMPLE-Mode - Jump to bootloader main BL-loop if key is 
     pressed (Pin grounded) "during" reset or jump to the 
	 application if the pin is not grounded (=pulled up by
	 internal pull-up-resistor)
   * POWERSAVE-Mode - Startup is separated in two loops
     which makes power-saving a little easier if no firmware
	 is on the chip. Needs more memory
   * BOOTICE-Mode - to flash the JTAGICE upgrade.ebn file.
     No startup-sequence in this mode. Jump directly to the
     parser-loop on reset 
     XTAL in BOOTICEMODE must be 7372800 Hz to be compatible
     with the org. JTAGICE-Firmware */
#define START_SIMPLE
#define XTAL 16000000UL

// UART handling - some definitions from P. Fleury's Library - thanks
//#define BAUDRATE 57600
#define BAUDRATE 57600
#include "uart.h"

/* enable/disable write of lock-bits 
  WARNING: lock-bits can not be reseted by bootloader (as far as I know)
  Only protection no unprotection, "chip erase" from bootloader only
  clears the flash but does no real "chip erase" (this is not possible
  with a bootloader as far as I know)
  Keep this undefined!
 */
// #define WRITELOCKBITS

/* Select Boot Size (select one, comment out the others) */
//#define _B1024 
#define _B4096

#include "chipdef.h"

#define UART_RX_BUFFER_SIZE SPM_PAGESIZE
unsigned char gBuffer[UART_RX_BUFFER_SIZE];

#define my_eeprom_busy_wait()         do{}while(!eeprom_is_ready())

unsigned char BufferLoad(unsigned int , unsigned char ) ;
void BlockRead(unsigned int , unsigned char ) ;
void lcd_init(void);
u08 print_string(char* string, u08 len);
void next_line(void);
void delay_us(u16 num);
u08 delay_ms(u16 num, u08 check);
void clear_screen(void);

uint32_t address;
unsigned char device;

void send_boot(void)
{
	sendchar('P');
	sendchar('o');
	sendchar('l');
	sendchar('y');
	sendchar('B');
	sendchar('o');
	sendchar('t');
}

void (*jump_to_app)(void) = 0x0000;

int main(void)
{
  unsigned  tempi;
  char val;
  u16 count=0;

  cli();

  MCUSR = 0;
  wdt_disable();

  DDRA = 0;
  DDRB = 0;
  DDRC = 0;
  DDRD = 0;

  PORTD |= (1<<0) | (1<<1) | (1<<2) | (1<<3) | (1<<7);  
  
  MCUCR = (1<<7) | (1<<IVCE);       
  MCUCR = (1<<7) | (1<<IVSEL);             //move interrupt vectors to the Boot sector    
  
  BLDDR  &= ~(1<<BLPNUM);	// set as Input
  BLPORT |= (1<<BLPNUM);		// Enable pullup

  
  USART_Init(UART_BAUD_SELECT(BAUDRATE,XTAL),UARTSINGLE); 	// single speed

  //JS enable pullup on PB7, this checks if the download cable is plugged in
  PORTD |= (1<<2) | (1<<3) | (1<<7);  
  PORTD |= (1<<0); //enable pullup for RX0
  PORTB |= (1<<2);  
  PORTB |= (1<<3);  
  PORTB |= (1<<5);  
  PORTB |= (1<<6);  
  PORTB |= (1<<7);  //PB7 is the pin which is checked

  delay_us(1000);  //JS added for ATMega644

  if (PINB & 0x80) { //no cable plugged in
    MCUCR = (1<<7) | (1<<IVCE); 
    MCUCR = (1<<7) | (0<<IVSEL);             //move interruptvectors to the Application sector

    delay_us(1000);   //JS added for ATMega644
    jump_to_app();
  }

  lcd_init();
  clear_screen();

  print_string("Polybot Boot 1.1",16);
  next_line();

  char num='9';
  u08 start_program=0;

  while (count<11 && start_program==0) {
    for(u16 i=0;i<1000;i++) {
      if (!(BLPIN & (1<<BLPNUM))) { //start the program if SW1 is pressed
	MCUCR = (1<<7) | (1<<IVCE); 
	MCUCR = (1<<7) | (0<<IVSEL);             //move interruptvectors to the Application sector
	delay_us(1000);   //JS added for ATMega644
	jump_to_app();

      }
      
      if (UCSR0A & (1<<RXC0)) { //got a character
	start_program=1;
	break;
      }
      
      if (delay_ms(1,1) == 1) {
	start_program=1;
	break;
      }
    }
    count++;

    if (print_string(&num,1) == 1) {
      start_program=1;
      break;
    }
      
    num--;

  }

  if (start_program==0) {
    MCUCR = (1<<7) | (1<<IVCE); 
    MCUCR = (1<<7) | (0<<IVSEL);             //move interruptvectors to the Application sector
    delay_us(1000);
    jump_to_app();
  }

  //clear_screen();
  //print_string("Downloading...",14);


  //  if(BLPIN & (1<<BLPNUM))
  // {  
      // jump to main app if pin is not grounded
  //   MCUCR = (1<<IVCE); 
  //  MCUCR = (0<<IVSEL);             //move interruptvectors to the Application sector
  //  jump_to_app();	// Jump to application sector
  //}

  //JS turn on PolyBot LED
  DDRA |= (1<<4);
  PORTA |= (1<<4);

  for(;;)                             
    {   
      val=recchar();
      
      if(val=='a')                        //Autoincrement?
	{
	  turn_on_uart_transmit;
	  sendchar('Y');			//Autoincrement is quicker
	  turn_off_uart_transmit;
	}
      
      else if(val=='A')                   //write address 
	{
	  address=recchar();                //read address 8 MSB
	  address=(address<<8)|recchar();
	  
	  address=address<<1;               // !! convert from word address to byte address
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if(val=='b')
	{				
	  // Buffer load support
	  turn_on_uart_transmit;
	  sendchar('Y');					// Report buffer load supported
	  sendchar((UART_RX_BUFFER_SIZE >> 8) & 0xFF);
	  // Report buffer size in bytes
	  sendchar(UART_RX_BUFFER_SIZE & 0xFF);
	  turn_off_uart_transmit;
	}
      
      else if(val=='B')					// Start buffer load
	{
	  tempi = recchar() << 8;			// Load high byte of buffersize
	  tempi |= recchar();				// Load low byte of buffersize
	  val = recchar();				// Load memory type ('E' or 'F')

	  turn_on_uart_transmit;
	  sendchar (BufferLoad(tempi,val));
	  // Start downloading of buffer
	  turn_off_uart_transmit;
	  PORTA ^= (1<<4);
	}
      
      else if(val == 'g')					// Block read
	{
	  tempi = (recchar() << 8) | recchar();
	  
	  val = recchar();				// Get memtype
	  BlockRead(tempi,val);			// Perform the block read
	  PORTA ^= (1<<4); //blink LED
	}		
      
      else if(val=='e')                   //Chip erase 
	{
	  if (device == devtype)
	    {
	      // erase only main section (bootloader protection)
	      address = 0; 
	      while ( APP_END > address )
		//while ( 0x7c00 > address )
		{
		  boot_page_erase(address);	// Perform page erase
		  boot_spm_busy_wait();		// Wait until the memory is erased.
		  address += SPM_PAGESIZE;
		  PORTA ^= (1<<4); //blink LED
		}
	    }
	  boot_rww_enable();
	  turn_on_uart_transmit;
	  sendchar('\r');       
	  turn_off_uart_transmit;
	}
      
      else if(val=='E')                   //Exit upgrade
	{
	  wdt_enable(WDTO_15MS); // Enable Watchdog Timer to give reset
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if(val=='P')     // Enter programming mode 
	{
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if(val=='L')   // Leave programming mode
	{ 
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if (val=='p')		// return programmer type
	{
	  turn_on_uart_transmit;
	  sendchar('S');		// always serial programmer
	  turn_off_uart_transmit;
	} 
      
      else if(val=='t')                   // Return device type 
	{
	  turn_on_uart_transmit;
	  sendchar(devtype);
	  sendchar(0);
	  turn_off_uart_transmit;
	}
      
      else if ((val=='x')||(val=='y'))	// clear and set LED ignored
	{
	  recchar();
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if (val=='T')					// set device
	{
	  device = recchar();
	  turn_on_uart_transmit;
	  sendchar('\r');
	  turn_off_uart_transmit;
	}
      
      else if (val=='S')                  // Return software identifier 
	{
	  turn_on_uart_transmit;
	  send_boot();
	  turn_off_uart_transmit;
	}                
      
      else if (val=='V')                  // Return Software Version
	{
	  turn_on_uart_transmit;
	  sendchar(VERSION_HIGH);
	  sendchar(VERSION_LOW);
	  turn_off_uart_transmit;
	}        
      
      else if (val=='s')                  // Return Signature Byte
	{
	  turn_on_uart_transmit;
	  sendchar(sig_byte1);
	  sendchar(sig_byte2);
	  sendchar(sig_byte3);
	  turn_off_uart_transmit;
	}       
      
      else if(val != 0x1b) /* ESC */
	{
	  turn_on_uart_transmit;
	  sendchar('?');
	  turn_off_uart_transmit;
	}
    }	// end of "parser" for-loop
  
  return 0;
}


unsigned char BufferLoad(unsigned int size, unsigned char mem)
{
  unsigned int data, cnt;
  uint32_t tempaddress;
  
  for (cnt=0; cnt<UART_RX_BUFFER_SIZE; cnt++) {
    if (cnt<size) gBuffer[cnt]=recchar();
    else gBuffer[cnt]=0xFF;
  }
  cnt=0;
  
  tempaddress = address; // Store address in page
  
  my_eeprom_busy_wait();
  
  if (device == devtype)
    {
      if (mem == 'F')		// Flash
	{
	  do {
	    data=gBuffer[cnt++];
	    data|=(gBuffer[cnt++]<<8);

	    if (address < APP_END) {
	      boot_page_fill(address,data); //call asm routine. 
	      address=address+2;  	    // Select next word in memory
	    }
	    size -= 2; // Reduce number of bytes to write by two    
	  } while(size); // Loop until all bytes written
	  
	  //JS
	  //if (tempaddress < APP_END) {
	    boot_page_write(tempaddress);
	    boot_spm_busy_wait();
	    //}	
	  boot_rww_enable(); //Re-enable the RWW section

	  //while(1) {
	  //DDRA &= ~(1<<4);
	  //PORTA |= (1<<4);
	  //} //test code
	  
	} // End FLASH
      
      if (mem == 'E') // Start EEPROM
        {
	  address>>=1;
	  do {
	    EEARL = address;			// Setup EEPROM address
	    EEARH = (address >> 8);
	    address++;					// Select next byte
	    EEDR=gBuffer[cnt++];
	    
	    EECR |= (1<<EEMPE);			// Write data into EEPROM
	    EECR |= (1<<EEPE);
	    
	    while (EECR & (1<<EEPE));	// Wait for EEPROM write to finish
	    size--; // Decrease number of bytes to write
	  } while(size); // Loop until all bytes written
	  
	}
      
      return '\r'; // Report programming OK
      
    }
  
  return 0; // Report programming failed
}

void BlockRead(unsigned int size, unsigned char mem)
{
  unsigned int data;
  
  my_eeprom_busy_wait();
  
  if (mem == 'E')							// Read EEPROM
    {
      // address>>=1; // not needed here - hmm, somehow inconsistent TODO
      do {
	EEARL = address;				// Setup EEPROM address
	EEARH = (address >> 8);
	address++;						// Select next EEPROM byte
	EECR |= (1<<EERE);				// Read EEPROM
	sendchar(EEDR);					// Transmit EEPROM data to PC
	size--;							// Decrease number of bytes to read
      } while (size);					// Repeat until all block has been read
    }
  else									// Read Flash
    {
      do {
#if defined(RAMPZ)
	data = pgm_read_word_far(address);
#else
	data = pgm_read_word_near((uint16_t)address);
#endif
	sendchar((unsigned char)data);		//send LSB
	sendchar((unsigned char)(data >> 8));	//send MSB  
	address += 2;  					// Select next word in memory
	size -= 2;						// Subtract two bytes from number of bytes to read
      } while (size);					// Repeat until all block has been read
    }
}

void delay_us(u16 num) {  //approximate delay in microseconds
  u16 i;

  for (i=0;i<num;i++) {
    _delay_loop_2(4);
  }
}


void write_lcd(u08 data) {
  DDRC = 0xff;  //make port C outputs

  PORTC = data;
  delay_us(1);
  sbi(PORTB,0);  //set the clock high for the LCD 
  delay_us(1);
  cbi(PORTB,0);  //set the clock low for the LCD 
  delay_us(1);

  DDRC = 0x00;  //make port C inputs again
}

void write_control(u08 data) {
  cbi(PORTB,1); //set RS low
  write_lcd(data);
}

void write_data(u08 data) {
  sbi(PORTB,1); //set RS high
  write_lcd(data);
}

void clear_screen(void) {
  write_control(0x01);  //clear display
  delay_us(3400);
}

u08 delay_ms(u16 num, u08 check) {  //approximate delay in milliseconds
  u16 i;

  for (i=0;i<num*4000;i++) {
    _delay_loop_2(1);

    if ((UCSR0A & (1<<RXC0)) && check) { //got a character
      return 1;
    }
  }

  return 0;
}

void lcd_init(void) {
  DDRC = 0x00;

  sbi(DDRD,4);
  sbi(DDRD,5);
  sbi(DDRD,6);

  sbi(PORTD,4);
  cbi(PORTD,5);
  cbi(PORTD,6);

  sbi(DDRB,0);
  sbi(DDRB,1);
  cbi(PORTB,0); //set E low
  cbi(PORTB,1); //set RS low

  //initialize the LCD as described in the HD44780 datasheet

  write_control(0x38);  //function set
  delay_ms(5,0);

  write_control(0x38);  //function set
  delay_us(160);

  write_control(0x38);  //function set
  delay_us(160);
  write_control(0x38);  //function set
  delay_us(160);
  write_control(0x08);  //turn display off
  delay_us(160);
  write_control(0x01);  //clear display
  delay_us(4000);
  write_control(0x06);  //set entry mode
  delay_us(160);

  write_control(0x0C); //turn on lcd
  delay_us(160);
}

u08 print_string(char* string, u08 len) {
  u08 i;

  for (i=0;i<len;i++) {
    write_data(string[i]);

    if (UCSR0A & (1<<RXC0)) { //got a character
      return 1;
    }
    delay_us(100);
  }

  return 0;
}

void next_line(void) {
  write_control(0xc0);  //go to the second line on the display
  delay_us(100);
}
